# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import collections
import netaddr
from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
from oslo_utils import versionutils

from neutron.common import utils as n_utils

from networking_bagpipe.bagpipe_bgp.common import config
from networking_bagpipe.bagpipe_bgp.common import dataplane_utils
from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import net_utils
from networking_bagpipe.bagpipe_bgp import constants as consts
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn import identifier_allocators
from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance

from neutron.agent.common import ovs_lib

from neutron_lib.plugins.ml2 import ovs_constants as ovs_const

# man ovs-ofctl /32768
DEFAULT_OVS_FLOW_PRIORITY = 0x8000

# we want to avoid having our flows having a lowest priority than
# the default
DEFAULT_RULE_PRIORITY = DEFAULT_OVS_FLOW_PRIORITY + 0x1000

# priorities for IP match flows
# highest priority MAX_PREFIX_PRIORITY for MAX_PREFIX_LENGTH prefix
# (MAX_PREFIX_PRIORITY - MAX_PREFIX_LENGTH) for a zero length prefix
MAX_PREFIX_PRIORITY = DEFAULT_RULE_PRIORITY
MAX_PREFIX_LENGTH = 0x80  # 128

# fallback flows get a priority even lower
# (we round it for better readability of flow dumps)
FALLBACK_PRIORITY = MAX_PREFIX_PRIORITY - MAX_PREFIX_LENGTH - 0x80

NO_MPLS_PHY_INTERFACE = -1

VXLAN_TUNNEL = "vxlan"

OVS_DUMP_FLOW_FILTER = "| grep -v NXST_FLOW | perl -pe '"               \
    "s/ *cookie=[^,]+, duration=[^,]+, table=[^,]+, //;" \
    "s/ *n_bytes=[^,]+, //; "                            \
    "s/ *(hard|idle)_age=[^,]+, //g; "                   \
    "s/n_packets=([0-9]),/packets=$1    /; "             \
    "s/n_packets=([0-9]{2}),/packets=$1   /; "           \
    "s/n_packets=([0-9]{3}),/packets=$1  /; "            \
    "s/n_packets=([0-9]+),/packets=$1 /; "               \
    "'"

GATEWAY_MAC = "00:00:5e:00:43:64"

ARP_RESPONDER_ACTIONS = ('move:NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[],'
                         'mod_dl_src:%(mac)s,'
                         'load:0x2->NXM_OF_ARP_OP[],'
                         'move:NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[],'
                         'push:NXM_OF_ARP_TPA[],'
                         'move:NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[],'
                         'load:%(mac)#x->NXM_NX_ARP_SHA[],'
                         'pop:NXM_OF_ARP_SPA[],'
                         '%(vlan_action)soutput:%(in_port)d')

VRF_REGISTER = 0


def _match_from_prefix(prefix):
    # A zero-length prefix is a default route, no nw_dst is needed/possible
    # in this case
    prefix_length = netaddr.IPNetwork(prefix).prefixlen
    return {'nw_dst': prefix} if prefix_length != 0 else {}


def _priority_from_prefix(prefix):
    # FIXME: use a priority depending on the prefix len
    #        to compensate the fact that "OpenFlow  leaves  behavior
    #        undefined when two or more flows with the same priority
    #        can match a single packet.  Some users expect ``sensible''
    #        behavior, such as more specific flows taking precedence
    #        over less specific flows, but OpenFlow does not specify
    #        this and Open vSwitch does not implement it.  Users should
    #        therefore  take  care  to  use  priorities  to ensure the
    #        behavior that they expect.
    prefix_length = netaddr.IPNetwork(prefix).prefixlen
    # to implement a longest-match lookup we give longest prefixes
    # the higher priority
    priority = MAX_PREFIX_PRIORITY - (MAX_PREFIX_LENGTH - prefix_length)
    return priority


class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):

    def __init__(self, *args, **kwargs):
        super(MPLSOVSVRFDataplane, self).__init__(*args, **kwargs)

        # Initialize dict where we store info on OVS ports (port numbers and
        # bound IP address)
        self._ovs_port_info = dict()

        self.bridge = self.driver.bridge

        self.nh_group_mgr = NextHopGroupManagerProxy(
            self.driver.nh_group_mgr,
            self.instance_id,
            self.driver.vrf_table,
            self._vrf_match(),
            self._cookie
        )

        self.fallback = None
        self.ovs_vlan = None

    @log_decorator.log_info
    def cleanup(self):
        if self._ovs_port_info:
            self.log.warning("OVS port numbers list for local ports plugged in"
                             " VRF is not empty, clearing...")
            self._ovs_port_info.clear()

        # Remove all flows for this instance
        for table in self.driver.all_tables.values():
            self.bridge.delete_flows(table=table,
                                     cookie=self._cookie(add=False))

        # Remove all groups for this instance
        self.nh_group_mgr.clear_objects()

    @log_decorator.log
    def _extract_mac_address(self, output):
        """Extract MAC address from command output"""

        return re.search(r"([0-9A-F]{2}[:-]){5}([0-9A-F]{2})", output,
                         re.IGNORECASE).group()

    def _find_remote_mac_address(self, remote_ip):
        """Find MAC address for a remote IP address"""

        # PING remote IP address
        (_, exit_code) = self._run_command("fping -r4 -t100 -q -I %s %s" %
                                           (self.bridge.br_name, remote_ip),
                                           raise_on_error=False,
                                           acceptable_return_codes=[-1])
        if exit_code != 0:
            self.log.info("can't ping %s via %s, proceeding anyways",
                          remote_ip, self.bridge.br_name)
            # we proceed even if the ping failed, since the ping was
            # just a way to trigger an ARP resolution which may have
            # succeeded even if the ping failed

        # Look in ARP cache to find remote MAC address
        (output, _) = self._run_command("ip neigh show to %s dev %s" %
                                        (remote_ip, self.bridge.br_name))

        if (not output or
                "FAILED" in output[0] or
                "INCOMPLETE" in output[0]):
            raise exc.RemotePEMACAddressNotFound(remote_ip)

        try:
            return self._extract_mac_address(output[0])
        except Exception:
            raise exc.RemotePEMACAddressNotFound(remote_ip)

    def _mtu_fixup(self, localport):
        # This is a hack, proper MTUs should actually be configured in the
        # hybrid vif driver
        # TODO(tmorin): obsolete

        mtu = self.config["ovsbr_interfaces_mtu"]
        if not mtu:
            self.log.debug("No ovsbr_interfaces_mtu specified in config file,"
                           " not trying to fixup MTU")
            return

        try:
            itf = localport['ovs']['port_name']
        except KeyError:
            self.log.warning("No OVS port name provided, cannot fix MTU")
            return

        self.log.info("Will adjust %s if with MTU %s "
                      "(ovsbr_interfaces_mtu specified in config)", itf, mtu)

        (_, exit_code) = self._run_command("ip link show %s" % itf,
                                           raise_on_error=False,
                                           acceptable_return_codes=[0, 1])

        if exit_code != 0:
            self.log.warning("No %s if, not trying to fix MTU", itf)
        else:
            self._run_command("ip link set %s mtu %s" % (itf, mtu),
                              run_as_root=True)

    def _get_ovs_port_specifics(self, localport):
        # Returns a tuple of:
        # - OVS port number for traffic to/from VMs
        # - OVS port name
        try:
            port_name = ""
            if ('ovs' in localport and localport['ovs']['plugged']):
                try:
                    port = localport['ovs']['port_number']
                except KeyError:
                    self.log.info("No OVS port number provided, trying to use"
                                  " a port name")
                    port = self.driver.find_ovs_port(
                        localport['ovs']['port_name'])
            else:
                try:
                    try:
                        port_name = localport['ovs']['port_name']
                    except KeyError:
                        port_name = localport['linuxif']
                except Exception:
                    raise Exception("Trying to find which port to plug, but no"
                                    " portname was provided")

                try:
                    port = self.driver.find_ovs_port(port_name)
                except Exception:
                    port = self.bridge.add_port(port_name)
                self.log.debug("Corresponding port number: %s", port)
        except KeyError as e:
            self.log.error("Incomplete port specification: %s", e)
            raise Exception("Incomplete port specification: %s" % e)

        return (port, port_name)

    def _vlan_match(self):
        return {'dl_vlan': self.ovs_vlan} if self.ovs_vlan else {}

    def get_push_vlan_action(self):
        return ("push_vlan:0x8100,mod_vlan_vid:%d" % self.ovs_vlan
                if self.ovs_vlan else "")

    def get_strip_vlan_action(self):
        return "strip_vlan" if self.ovs_vlan else ""

    @log_decorator.log_info
    def update_fallback(self, fallback=None):
        if fallback:
            self.fallback = fallback

        if not self.fallback:
            return

        for param in ('src_mac', 'dst_mac', 'ovs_port_number'):
            if not self.fallback.get(param):
                self.log.error("fallback specified without '%s'", param)
                return

        # use priority -1 so that the route only hits when the packet
        # does not matches any VRF route
        self.bridge.add_flow_extended(
            flow_matches=[dict(table=self.driver.vrf_table,
                               cookie=self._cookie(add=True),
                               priority=FALLBACK_PRIORITY),
                          self._vrf_match()],
            actions=[self.get_push_vlan_action(),
                     'mod_dl_src:%s' % self.fallback.get('src_mac'),
                     'mod_dl_dst:%s' % self.fallback.get('dst_mac'),
                     'output:%d' % self.fallback.get('ovs_port_number')])

    @log_decorator.log_info
    def setup_arp_responder(self, ovs_port):
        if not self.driver.config.arp_responder:
            return

        self.bridge.add_flow_extended(
            flow_matches=[dict(table=self.driver.vrf_table,
                               cookie=self._cookie(add=True),
                               priority=DEFAULT_RULE_PRIORITY,
                               proto='arp',
                               dl_dst='ff:ff:ff:ff:ff:ff',
                               arp_op='0x1'),
                          self._vrf_match(),
                          # Respond to all IP addresses if proxy ARP is
                          # enabled, otherwise only for gateway
                          {'arp_tpa': self.gateway_ip}
                          if not self.driver.config.proxy_arp else {}],
            actions=[ARP_RESPONDER_ACTIONS % {
                'mac': netaddr.EUI(GATEWAY_MAC, dialect=netaddr.mac_unix),
                'vlan_action': self.get_push_vlan_action(),
                'in_port': ovs_port
            }])

    @log_decorator.log_info
    def remove_arp_responder(self):
        self.bridge.delete_flows_extended(
            flow_matches=[dict(table=self.driver.vrf_table,
                               cookie=self._cookie(add=False),
                               proto='arp'),
                          self._vrf_match()])

    def _check_vlan_use(self, localport):
        # checks that if a vlan_action is used, it is the same
        # for all interfaces plugged into the VRF
        # on first plug we update
        try:
            ovs_vlan = int(localport['ovs']['vlan'])
        except KeyError:
            return

        if self.ovs_vlan is None:
            self.ovs_vlan = ovs_vlan
        else:
            # on a subsequent plug, we check
            if self.ovs_vlan != ovs_vlan:
                self.log.error("different VLAN for different interfaces: "
                               "%s vs %s", self.ovs_vlan,
                               ovs_vlan)
                raise Exception("can't specify a different VLAN for different"
                                " interfaces")

    @log_decorator.log
    def vif_plugged(self, mac_address, ip_address, localport, label,
                    direction):

        (ovs_port, ovs_port_name) = self._get_ovs_port_specifics(localport)

        self._check_vlan_use(localport)

        # need to update fallback, in case it was called before the
        # first vifPlugged call could define the push_vlan action
        self.update_fallback()

        # This is a hack used with previous versions of Openstack
        #  proper MTUs should actually be configured in the hybrid vif driver
        # Please consider this obsolete until it gets clean'd up
        self._mtu_fixup(localport)

        if vpn_instance.forward_from_port(direction):
            # Map traffic from plugged port to VRF

            # Note that we first reset the in_port so that OVS will allow the
            # packet to eventually go back to this port after a VRF lookup
            # (the case where that happens is where in port is a patch-port on
            # which different VLANs are used to reach different networks):
            # patch port vlan X -- VRFX --- VRF Y -- patch-port vlan Y
            #
            # ( see http://docs.openvswitch.org/en/latest/faq/openflow/
            # "Q: I added a flow to send packets out the ingress port..." )
            for proto in ('ip', 'arp'):
                self.bridge.add_flow_extended(
                    flow_matches=[dict(table=self.driver.input_table,
                                       cookie=self._cookie(add=True),
                                       priority=DEFAULT_RULE_PRIORITY,
                                       proto=proto,
                                       in_port=ovs_port),
                                  self._vlan_match()],
                    actions=[self.get_strip_vlan_action(),
                             'load:0->NXM_OF_IN_PORT[],',
                             'set_field:%d->reg%d,' % (self.instance_id,
                                                       VRF_REGISTER),
                             'resubmit(,%d)' % self.driver.vrf_table])

        # Map ARP responder if necessary
        if not self._ovs_port_info:
            self.setup_arp_responder(ovs_port)

        if vpn_instance.forward_to_port(direction):
            # Map incoming MPLS traffic going to the VM port
            incoming_actions = [self.get_push_vlan_action(),
                                "mod_dl_src:%s,mod_dl_dst:%s" % (GATEWAY_MAC,
                                                                 mac_address),
                                "output:%s" % ovs_port]

            self.bridge.add_flow_extended(
                flow_matches=[dict(table=self.driver.encap_in_table,
                                   cookie=self._cookie(add=True),
                                   priority=DEFAULT_RULE_PRIORITY,
                                   proto="mpls",
                                   mpls_label=label,
                                   mpls_bos=1)],
                actions=["pop_mpls:0x0800"] + incoming_actions)

            # additional incoming traffic rule for VXLAN
            if self.driver.vxlan_encap:
                self.bridge.add_flow_extended(
                    flow_matches=[
                        dict(table=self.driver.encap_in_table,
                             cookie=self._cookie(add=True),
                             priority=DEFAULT_RULE_PRIORITY,
                             in_port=self.driver.vxlan_tunnel_port_number,
                             tun_id=label)],
                    actions=incoming_actions)

        # Add OVS port number in list for local port plugged in VRF
        # FIXME: check check check, is linuxif the right key??
        self.log.debug("Adding OVS port %s with port %s for address "
                       "%s, to the list of ports plugged in VRF",
                       localport['linuxif'], ovs_port, ip_address)
        self._ovs_port_info[localport['linuxif']] = {
            "ovs_port": ovs_port,
            "ovs_port_name": ovs_port_name,
        }

    @log_decorator.log
    def vif_unplugged(self, mac_address, ip_address, localport, label,
                      direction, last_endpoint=True):

        ovs_port = self._ovs_port_info[localport['linuxif']]['ovs_port']
        ovs_port_name = self._ovs_port_info[
            localport['linuxif']]['ovs_port_name']

        if vpn_instance.forward_to_port(direction):
            # Unmap incoming MPLS traffic going to the VM port
            self.bridge.delete_flows(table=self.driver.encap_in_table,
                                     cookie=self._cookie(add=False),
                                     proto="mpls",
                                     mpls_label=label,
                                     mpls_bos=1)

            # Unmap incoming VXLAN traffic...
            if self.driver.vxlan_encap:
                self.bridge.delete_flows(
                    table=self.driver.encap_in_table,
                    cookie=self._cookie(add=False),
                    in_port=self.driver.vxlan_tunnel_port_number,
                    tun_id=label)

        if last_endpoint:
            if vpn_instance.forward_from_port(direction):
                # Unmap all traffic from plugged port
                self.bridge.delete_flows_extended(
                    flow_matches=[dict(table=self.driver.input_table,
                                       cookie=self._cookie(add=False),
                                       in_port=ovs_port),
                                  self._vlan_match()])

            # Unmap ARP responder
            self.remove_arp_responder()

            # Run port unplug action if necessary (OVS port delete)
            if ovs_port_name:
                self.bridge.delete_port(ovs_port_name)

            # Remove OVS port number from list for local port plugged in VRF
            del self._ovs_port_info[localport['linuxif']]

    def _get_label_action(self, label, encaps):
        if (self.driver.vxlan_encap and
                exa.Encapsulation(exa.Encapsulation.Type.VXLAN) in encaps):
            return "set_field:%d->tunnel_id" % label
        else:
            return "push_mpls:0x8847,set_mpls_label:%d" % label

    def _get_output_action(self, remote_pe, encaps):
        # Check if prefix is from a local VRF
        if self.driver.get_local_address() == remote_pe:
            self.log.debug("Local route, using a resubmit action")
            # For local traffic, we have to use a resubmit action
            if (self.driver.vxlan_encap and
                    exa.Encapsulation(exa.Encapsulation.Type.VXLAN) in encaps):
                return ("resubmit(%d,%d)" %
                        (self.driver.vxlan_tunnel_port_number,
                         self.driver.encap_in_table))
            else:
                return "resubmit(%d,%d)" % (self.driver.mpls_in_port(),
                                            self.driver.encap_in_table)
        else:
            if (self.driver.vxlan_encap and
                    exa.Encapsulation(exa.Encapsulation.Type.VXLAN) in encaps):
                self.log.debug("Will use a VXLAN encap for this destination")
                return "set_field:%s->tun_dst,output:%s" % (
                    remote_pe, self.driver.vxlan_tunnel_port_number)
            elif self.driver.use_gre:
                self.log.debug("Using MPLS/GRE encap")
                return "set_field:%s->tun_dst,output:%s" % (
                    remote_pe, self.driver.gre_tunnel_port_number)
            else:
                self.log.debug("Using bare MPLS encap")
                # Find remote router MAC address
                try:
                    remote_pe_mac_address = self._find_remote_mac_address(
                        remote_pe)
                    self.log.debug("MAC address found for remote router "
                                   "%(remote_pe)s: %(remote_pe_mac_address)s",
                                   locals())
                except exc.RemotePEMACAddressNotFound as e:
                    self.log.error("An error occured during setupDataplaneFor"
                                   "RemoteEndpoint: %s", e)
                    raise

                # Map traffic to remote IP address as MPLS on ethX to remote
                # router MAC address
                return "mod_dl_src:%s,mod_dl_dst:%s,output:%s" % (
                    self.driver.mpls_if_mac_address, remote_pe_mac_address,
                    self.driver.ovs_mpls_if_port_number)

    def _cookie(self, add=False):
        mask = ""
        if not add:
            mask = "/-1"
        return "%d%s" % (self.instance_id, mask)

    def _vrf_match(self):
        return {'reg%d' % VRF_REGISTER: self.instance_id}

    @log_decorator.log_info
    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                            nlri, encaps,
                                            lb_consistent_hash_order=0):
        nexthop = NextHop(label, remote_pe, encaps, lb_consistent_hash_order)

        if self.nh_group_mgr.is_object_user(prefix, nexthop):
            self.log.debug("Dataplane already in place for %s, %s, skipping",
                           prefix, nexthop)
            return

        dec_ttl = (netaddr.IPNetwork(prefix) not in netaddr.IPNetwork(
                   "%s/%s" % (self.gateway_ip, self.network_plen)))
        label_action = self._get_label_action(nexthop.label,
                                              nexthop.encaps)
        output_action = self._get_output_action(nexthop.remote_pe,
                                                nexthop.encaps)
        self.nh_group_mgr.new_nexthop(prefix, nexthop,
                                      actions=["dec_ttl" if dec_ttl else "",
                                               label_action,
                                               output_action])

    @log_decorator.log_info
    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                             nlri, encaps,
                                             lb_consistent_hash_order=0):

        nexthop = NextHop(label, remote_pe, encaps, lb_consistent_hash_order)

        if self.nh_group_mgr.is_object_user(prefix, nexthop):
            self.nh_group_mgr.del_nexthop(prefix, nexthop)
        else:
            self.log.debug("remove_dataplane_for_remote_endpoint called, "
                           "for %s and %s, but we don't know about this "
                           "prefix and next-hop", prefix, nexthop.__dict__)

    def _get_port_range_from_classifier(self, classifier_port):
        if classifier_port:
            if isinstance(classifier_port, tuple):
                port_min, port_max = classifier_port
            else:
                port_min = port_max = classifier_port

        return port_min, port_max

    def _create_port_range_flow_matches(self, classifier_match, classifier):
        flow_matches = []
        src_port_match = '{:s}_src'.format(classifier.protocol)

        if classifier.source_port:
            if isinstance(classifier.source_port, tuple):
                src_port_min, src_port_max = classifier.source_port
            else:
                src_port_min = src_port_max = classifier.source_port

        dst_port_match = '{:s}_dst'.format(classifier.protocol)

        if classifier.destination_port:
            if isinstance(classifier.destination_port, tuple):
                dst_port_min, dst_port_max = classifier.destination_port
            else:
                dst_port_min = dst_port_max = classifier.destination_port

        dst_port_range = []
        if dst_port_min and dst_port_max:
            dst_port_range = n_utils.port_rule_masking(dst_port_min,
                                                       dst_port_max)
        src_port_range = []
        if src_port_min and src_port_max:
            src_port_range = n_utils.port_rule_masking(src_port_min,
                                                       src_port_max)
            for port in src_port_range:
                flow_match = classifier_match.copy()
                flow_match[src_port_match] = port
                if dst_port_range:
                    for port in dst_port_range:
                        dst_flow = flow_match.copy()
                        dst_flow[dst_port_match] = port
                        flow_matches.append(dst_flow)
                else:
                    flow_matches.append(flow_match)
        else:
            for port in dst_port_range:
                flow_match = classifier_match.copy()
                flow_match[dst_port_match] = port
                flow_matches.append(flow_match)

        return flow_matches

    def _create_classifier_flow_matches(self, classifier):
        classifier_match = dict(proto=classifier.protocol)

        if classifier.source_pfx:
            classifier_match.update({'nw_src': classifier.source_pfx})

        if classifier.destination_pfx:
            classifier_match.update({'nw_dst': classifier.destination_pfx})

        return self._create_port_range_flow_matches(classifier_match,
                                                    classifier)

    @log_decorator.log_info
    def add_dataplane_for_traffic_classifier(self, classifier,
                                             redirect_to_instance_id):
        classifier_matches = self._create_classifier_flow_matches(classifier)

        # Add traffic redirection to redirection VRF for classifier matches
        for classifier_match in classifier_matches:
            self.bridge.add_flow_extended(
                flow_matches=[dict(table=self.driver.vrf_table,
                                   cookie=self._cookie(add=True),
                                   priority=DEFAULT_RULE_PRIORITY),
                              self._vrf_match(),
                              classifier_match],
                actions=['set_field:%d->reg%d' % (redirect_to_instance_id,
                                                  VRF_REGISTER),
                         'resubmit(,%d)' % self.driver.vrf_table])

    @log_decorator.log_info
    def remove_dataplane_for_traffic_classifier(self, classifier):
        classifier_matches = self._create_classifier_flow_matches(classifier)

        # Remove traffic redirection to redirection VRF for classifier matches
        for classifier_match in classifier_matches:
            self.bridge.delete_flows_extended(
                flow_matches=[dict(table=self.driver.vrf_table,
                                   cookie=self._cookie(add=False)),
                              self._vrf_match(),
                              classifier_match])

    def get_lg_map(self):
        return {
            "flows": (lg.SUBTREE, self.get_lg_ovs_flows)
        }

    def get_lg_ovs_flows(self, path_prefix):
        return self.driver.get_lg_ovs_flows(
            path_prefix, 'cookie=%s' % self._cookie(add=False))


class OVSGroupAllocator(identifier_allocators.IDAllocator):

    MAX = 2 ** 32 - 1


class OVSBucketAllocator(identifier_allocators.IDAllocator):

    # Values greater than 0xFFFFFF00 are reserved
    MAX = 2 ** 32 - 2 ** 8 - 1


class NextHop(object):

    def __init__(self, label, remote_pe, encaps, lb_consistent_hash_order):
        self.label = label
        self.remote_pe = str(remote_pe)
        self.encaps = frozenset(encaps)
        self.lb_consistent_hash_order = lb_consistent_hash_order

    def __eq__(self, other):
        return ((self.label, self.remote_pe, self.encaps) ==
                (other.label, other.remote_pe, other.encaps))

    def __hash__(self):
        return hash((self.label,
                     self.remote_pe,
                     self.encaps))

    def __repr__(self):
        return "NextHop(%s,%s,%s,%s)" % (self.label,
                                         self.remote_pe,
                                         self.encaps,
                                         self.lb_consistent_hash_order)


class NextHopGroupManager(dataplane_utils.ObjectLifecycleManager):

    def __init__(self, bridge, hash_method, hash_method_param, hash_fields):
        super(NextHopGroupManager, self).__init__()

        self.bridge = bridge
        self.hash_method = hash_method
        self.hash_method_param = hash_method_param
        self.hash_fields = hash_fields

        self.group_allocator = OVSGroupAllocator()

    def get_selection_method(self):
        selection_method = self.hash_method
        if self.hash_fields and self.hash_method == 'hash':
            selection_method += ",fields(%s)" % ','.join(self.hash_fields)

        return selection_method

    @log_decorator.log_info
    def create_object(self, prefix, *args, **kwargs):
        buckets = (
            {'buckets': kwargs['buckets']} if kwargs.get('buckets') else {}
        )

        group_id = self.group_allocator.get_new_id("Group ID for prefix %s" %
                                                   str(prefix))

        self.bridge.add_group(group_id=group_id,
                              type='select',
                              selection_method=self.get_selection_method(),
                              selection_method_param=self.hash_method_param,
                              **buckets)

        return group_id

    @log_decorator.log_info
    def delete_object(self, group_id):
        self.bridge.delete_group(group_id=group_id)
        self.group_allocator.release(group_id)


class NextHopGroupManagerProxy(dataplane_utils.ObjectLifecycleManagerProxy):

    def __init__(self, manager, parent_key, vrf_table, vrf_match, cookie_func):
        super(NextHopGroupManagerProxy, self).__init__(manager, parent_key)

        self.vrf_table = vrf_table
        self.vrf_match = vrf_match
        self.cookie_func = cookie_func

        self.bucket_allocators = (
            collections.defaultdict(OVSBucketAllocator)
        )

        self.prefix_nexthop_2_bucket = dict()

    def _update_group_buckets(self, group_id, prefix):
        buckets = []
        for prefix_nh, bucket in self.prefix_nexthop_2_bucket.items():
            if prefix == prefix_nh[0]:
                buckets.append('bucket=bucket_id=%d,%s' % (
                    bucket[0], dataplane_utils.join_s(*bucket[1])))

        self.manager.bridge.mod_group(
            group_id=group_id,
            type='select',
            selection_method=self.manager.get_selection_method(),
            selection_method_param=self.manager.hash_method_param,
            buckets=','.join(buckets))

    def new_nexthop(self, prefix, nexthop, actions=None):
        if actions is None:
            actions = []
        bucket_allocator = self.bucket_allocators[prefix]
        bucket_id = bucket_allocator.get_new_id(
            "Bucket ID for prefix %s and nexthop %s" % (str(prefix), nexthop),
            hint_value=nexthop.lb_consistent_hash_order
        )
        bucket = 'bucket=bucket_id=%d,%s' % (bucket_id,
                                             dataplane_utils.join_s(*actions))

        self.prefix_nexthop_2_bucket[(prefix, nexthop)] = (bucket_id, actions)

        group_id, first = self.get_object(prefix, nexthop, buckets=bucket)
        if first:
            self.manager.bridge.add_flow_extended(
                flow_matches=[dict(table=self.vrf_table,
                                   cookie=self.cookie_func(add=True),
                                   priority=_priority_from_prefix(prefix),
                                   proto='ip'),
                              self.vrf_match,
                              _match_from_prefix(prefix)],
                actions=["group:%d" % group_id])
        else:
            self._update_group_buckets(group_id, prefix)

    def del_nexthop(self, prefix, nexthop):
        group_id = self.find_object(prefix)

        bucket_id, _ = self.prefix_nexthop_2_bucket.pop((prefix, nexthop))

        bucket_allocator = self.bucket_allocators[prefix]
        bucket_allocator.release(bucket_id)

        self._update_group_buckets(group_id, prefix)

        last = self.free_object(prefix, nexthop)

        if last:
            self.manager.bridge.delete_flows_extended(
                flow_matches=[dict(strict=True,
                                   table=self.vrf_table,
                                   cookie=self.cookie_func(add=False),
                                   priority=_priority_from_prefix(prefix),
                                   proto='ip'),
                              self.vrf_match,
                              _match_from_prefix(prefix)])

            del self.bucket_allocators[prefix]


class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):

    """Dataplane driver using OpenVSwitch

    Based on an OpenVSwitch 2.4 MPLS kernel dataplane implementation.

    This driver was successfully tested with the OVS 2.4 DKMS module.

    This driver uses MPLS-over-GRE by default. However, note well that current
    OVS implementation of MPLS-over-GRE is not yet conformant with RFC4023,
    because of an intermediate Eth header (MPLS-over-Eth-over-GRE).

    If MPLS-over-GRE is disabled (with mpls_over_gre=False), this driver
    currently requires that the OVS bridge be associated to the address used as
    the local_address in bgp.conf, to allow the linux IP stack to use the same
    physical interface as the one on which MPLS packets are forwarded. This
    requires to configure the OVS bridge so that it passes packets from the
    physical interface to the linux IP stack if they are not MPLS, and packets
    from the linux IP stack to the physical device.

    Howto allow the use of the OVS bridge interface also as an IP
    interface of the Linux kernel IP stack:
        ovs-ofctl del-flows br-int
        ovs-ofctl add-flow br-int in_port=LOCAL,action=output:1
        ovs-ofctl add-flow br-int in_port=1,action=output:LOCAL

    (on a debian or ubuntu system, this can be done part of the ovs bridge
    definition in /etc/network/interfaces, as post-up commands)

    The 'vrf_table' (resp. 'input_table') config parameters can be
    used to specify which OVS table will host the rules for traffic from VRFs
    (resp. for incoming traffic). Beware, this dataplane driver will
    *not* take care of setting up rules so that MPLS traffic or the traffic
    from attached ports is matched against rules in these tables.
    """

    dataplane_instance_class = MPLSOVSVRFDataplane
    type = consts.IPVPN
    ecmp_support = True
    required_ovs_version = "2.8.0"

    driver_opts = [
        cfg.StrOpt("mpls_interface",
                   help=("Interface used to send/receive MPLS traffic. "
                         "Use '*gre*' to choose automatic creation of a tunnel"
                         " port for MPLS/GRE encap")),
        cfg.StrOpt("mpls_over_gre",
                   choices=['auto', 'True', 'False'],
                   default="auto",
                   advanced=True,
                   help=("Force the use of MPLS/GRE even with "
                         "mpls_interface specified")),
        cfg.BoolOpt("proxy_arp", default=False,
                    advanced=True,
                    help=("Activate ARP responder per VRF for any IP "
                          "address")),
        cfg.BoolOpt("arp_responder", default=False,
                    advanced=True,
                    help=("ARP responder per VRF")),
        cfg.BoolOpt("vxlan_encap", default=False,
                    advanced=True,
                    help=("Be ready to receive VPN traffic as VXLAN, and to "
                          "preferrably send traffic as VXLAN when advertised "
                          "by the remote end")),
        cfg.StrOpt("ovs_bridge", default="br-mpls", advanced=True),
        cfg.IntOpt("input_table", default=0, advanced=True),
        cfg.IntOpt("ovs_table_id_start", default=1, advanced=True),
        cfg.StrOpt("gre_tunnel", default="mpls_gre", advanced=True,
                   help="OVS interface name for MPLS/GRE encap"),
        cfg.ListOpt("gre_tunnel_options", default=[],
                    item_type=types.String(),
                    help=("Options, comma-separated, passed to OVS for GRE "
                          "tunnel port creation (e.g. 'packet_type=legacy_l3"
                          ", ...') that will be added as OVS tunnel "
                          "interface options (e.g. 'options:packet_type="
                          "legacy_l3 options:...')")),
        cfg.IntOpt("ovsbr_interfaces_mtu", advanced=True),
        cfg.StrOpt("hash_method",
                   choices=["hash", "dp_hash"],
                   default="dp_hash",
                   advanced=True,
                   help=("Can be used to control the OVS group bucket "
                         "selection method (mapped to ovs "
                         "'selection_method')")),
        cfg.StrOpt("hash_method_param",
                   default=0,
                   advanced=True,
                   help=("Can be used to control the OVS group bucket "
                         "selection method (mapped to ovs "
                         "'selection_method_param')")),
        cfg.ListOpt("hash_fields", default=[], advanced=True,
                    help=("Can be used to control the fields used by the OVS "
                          "group bucket selection method (mapped to ovs "
                          "'fields')"))
    ]

    def __init__(self):
        super(MPLSOVSDataplaneDriver, self).__init__()

        config.set_default_root_helper()
        config.setup_privsep()

        try:
            (o, _) = self._run_command("ovs-ofctl -V | head -1 |"
                                       " awk '{print $4}'")
            self.ovs_release = o[0]
            self.log.info("OVS version: %s", self.ovs_release)
        except Exception:
            self.log.warning("Could not determine OVS release")
            self.ovs_release = None

        self.mpls_interface = self.config.mpls_interface

        if self.config.mpls_over_gre != "auto":
            self.use_gre = True
        else:
            self.use_gre = not (self.mpls_interface and
                                self.mpls_interface != "*gre*")

        if not self.mpls_interface:
            if not self.use_gre:
                raise Exception("mpls_over_gre force-disabled, but no "
                                "mpls_interface specified")
            else:
                self.use_gre = True
                self.log.info("Defaulting to use of MPLS-over-GRE (no "
                              "mpls_interface specified)")
        elif self.mpls_interface == "*gre*":
            if not self.use_gre:
                raise Exception("mpls_over_gre force-disabled, but "
                                "mpls_interface set to '*gre', cannot "
                                "use bare MPLS")
            else:
                self.log.info("mpls_interface is '*gre*', will thus use "
                              "MPLS-over-GRE")
                self.use_gre = True
                self.mpls_interface = None
        else:
            if self.use_gre:
                self.log.warning("mpls_over_gre set to True, "
                                 "ignoring mpls_interface parameter")
                self.mpls_interface = None
            else:
                self.log.info("Will use bare MPLS on interface %s",
                              self.mpls_interface)

        self.input_table = self.config.input_table

        if self.config.ovs_table_id_start == self.input_table:
            raise Exception("invalid ovs_table_id_start (%d): can't use tables"
                            " same as input table (%d)" % (
                                self.config.ovs_table_id_start,
                                self.config.input_table))

        self.encap_in_table = self.config.ovs_table_id_start
        self.vrf_table = self.config.ovs_table_id_start + 1

        self.all_tables = {'incoming': self.input_table,
                           'vrf': self.vrf_table,
                           'encap_in': self.encap_in_table}

        # Used to control whether this VRF will support
        # receiving traffic as VXLAN
        self.vxlan_encap = self.config.vxlan_encap

        # check OVS version
        ovs_release_version = versionutils.convert_version_to_tuple(
            self.ovs_release)
        required_ovs_version_tuple = versionutils.convert_version_to_tuple(
            self.required_ovs_version)
        if (not self.vxlan_encap and
                ovs_release_version < required_ovs_version_tuple):
            self.log.warning("%s requires at least OVS %s"
                             " (you are running %s)",
                             self.__class__.__name__,
                             self.required_ovs_version,
                             self.ovs_release)

        # unless useGRE is enabled, check that fping is installed
        if not self.use_gre:
            self._run_command("fping -v", raise_on_error=True)

        self.bridge = dataplane_utils.OVSBridgeWithGroups(
            dataplane_utils.OVSExtendedBridge(self.config.ovs_bridge)
        )
        # Check if OVS bridge exist
        if not self.bridge.bridge_exists(self.bridge.br_name):
            raise exc.OVSBridgeNotFound(self.bridge.br_name)

        self.bridge.use_at_least_protocol(ovs_const.OPENFLOW15)

        self.nh_group_mgr = NextHopGroupManager(self.bridge,
                                                self.config.hash_method,
                                                self.config.hash_method_param,
                                                self.config.hash_fields)

        if not self.use_gre:
            self.log.info("Will not force the use of GRE/MPLS, trying to bind "
                          "physical interface %s", self.mpls_interface)
            # Check if MPLS interface is attached to OVS bridge
            if not self.bridge.port_exists(self.mpls_interface):
                raise Exception("Specified mpls_interface %s is not plugged to"
                                " OVS bridge %s" % (self.mpls_interface,
                                                    self.bridge.br_name))
            else:
                self.ovs_mpls_if_port_number = self.bridge.get_port_ofport(
                    self.mpls_interface)

    def supported_encaps(self):
        if self.use_gre:
            yield exa.Encapsulation(exa.Encapsulation.Type.GRE)
            yield exa.Encapsulation(exa.Encapsulation.Type.DEFAULT)
            # we will accept routes with no encap
            # specified and force the use of GRE
        else:
            yield exa.Encapsulation(exa.Encapsulation.Type.MPLS)
            # we also accept route with no encap specified
            yield exa.Encapsulation(exa.Encapsulation.Type.DEFAULT)

        if self.vxlan_encap:
            yield exa.Encapsulation(exa.Encapsulation.Type.VXLAN)

    def mpls_in_port(self):
        if self.use_gre:
            return self.gre_tunnel_port_number
        else:
            return self.ovs_mpls_if_port_number

    @log_decorator.log_info
    def reset_state(self):
        # Flush all MPLS and ARP flows, all groups, if bridge exists

        if self.bridge.bridge_exists(self.bridge.br_name):
            self.log.info("Cleaning up OVS rules")

            self.bridge.delete_flows(table=self.input_table,
                                     cookie=ovs_lib.COOKIE_ANY,
                                     proto='mpls')
            if self.vxlan_encap:
                try:
                    self.bridge.delete_flows(
                        table=self.input_table,
                        in_port=self.find_ovs_port(VXLAN_TUNNEL))
                except Exception:
                    self.log.info("no VXLAN tunnel port, nothing to clean up")
                # the above won't clean up flows if the vxlan_tunnel interface
                # has changed...
                self.bridge.delete_flows(table=self.input_table,
                                         cookie=ovs_lib.COOKIE_ANY,
                                         tun_id='2/1')
                self.bridge.delete_flows(table=self.input_table,
                                         cookie=ovs_lib.COOKIE_ANY,
                                         tun_id='1/1')

            # clean input_table rule for plugged ports
            # NOTE(tmorin): would be cleaner using a cookie
            self.bridge.delete_flows(table=self.input_table,
                                     cookie=ovs_lib.COOKIE_ANY,
                                     proto='ip')
            self.bridge.delete_flows(table=self.input_table,
                                     cookie=ovs_lib.COOKIE_ANY,
                                     proto='arp')

            self.bridge.delete_flows(table=self.encap_in_table,
                                     cookie=ovs_lib.COOKIE_ANY)
            self.bridge.delete_flows(table=self.vrf_table,
                                     cookie=ovs_lib.COOKIE_ANY)

            # clean all groups
            self.bridge.delete_group()

            if self.log.isEnabledFor(logging.DEBUG):
                self.log.debug("All our rules have been flushed:\n%s",
                               '\n'.join(self.bridge.dump_all_flows()))
                self.log.debug("All groups have been flushed:\n%s",
                               self.bridge.run_ofctl("dump-groups", []))
        else:
            self.log.info("No OVS bridge (%s), no need to cleanup OVS rules",
                          self.bridge.br_name)

    def initialize(self):
        if self.use_gre:
            self.log.info("Setting up tunnel for MPLS/GRE (%s)",
                          self.config.gre_tunnel)

            self.bridge.delete_port(self.config.gre_tunnel)

            gre_tunnel_options = dict(
                o.split('=') for o in self.config.gre_tunnel_options)
            gre_tunnel_attrs = [
                ('type', 'gre'),
                ('options', dict({'local_ip': self.get_local_address(),
                                  'remote_ip': 'flow'},
                                 **gre_tunnel_options))
            ]

            self.gre_tunnel_port_number = (
                self.bridge.add_port(self.config.gre_tunnel, *gre_tunnel_attrs)
            )
            self.mpls_if_mac_address = None
        else:
            # Find ethX MPLS interface MAC address
            try:
                self.mpls_if_mac_address = net_utils.get_device_mac(
                    self._run_command,
                    self.mpls_interface)
            except Exception:
                # Interface without MAC address (patch port case), use MPLS
                # bridge MAC address instead
                self.mpls_if_mac_address = net_utils.get_device_mac(
                    self._run_command,
                    self.bridge.br_name)

        self.bridge.add_flow(table=self.input_table,
                             priority=DEFAULT_RULE_PRIORITY,
                             in_port=self.mpls_in_port(),
                             proto='mpls',
                             actions="resubmit(,%d)" % self.encap_in_table)

        if self.vxlan_encap:
            self.log.info("Enabling VXLAN encapsulation")
            self.bridge.delete_port(VXLAN_TUNNEL)

            vxlan_tunnel_attrs = [
                ('type', 'vxlan'),
                ('options', {'local_ip': self.get_local_address(),
                             'remote_ip': 'flow',
                             'key': 'flow'})
            ]
            self.vxlan_tunnel_port_number = (
                self.bridge.add_port(VXLAN_TUNNEL, *vxlan_tunnel_attrs)
            )

            self.bridge.add_flow(table=self.input_table,
                                 priority=DEFAULT_RULE_PRIORITY,
                                 in_port=self.vxlan_tunnel_port_number,
                                 actions="resubmit(,%d)" % self.encap_in_table)

    def validate_directions(self, direction):
        # this driver supports all combinations of directions
        pass

    def find_ovs_port(self, dev_name):
        """Find OVS port number from port name"""

        ofport = self.bridge.get_port_ofport(dev_name)

        if ofport is None:
            raise Exception("OVS port not found for device %s" % dev_name)

        return ofport

    # Looking glass code ####

    def get_lg_map(self):
        return {
            "flows": (lg.SUBTREE, self.get_lg_ovs_flows),
            "ports": (lg.SUBTREE, self.get_lg_ovs_ports)
        }

    def get_lg_local_info(self, path_prefix):
        d = {
            "ovs_bridge": self.bridge.br_name,
            "mpls_interface": self.mpls_interface,
            "gre": {'enabled': self.use_gre},
            "vxlan": {'enabled': self.vxlan_encap},
            "ovs_version": self.ovs_release,
            "tables": self.all_tables
        }

        if self.use_gre:
            d["gre"].update({'gre_tunnel_port': self.config.gre_tunnel})
        if self.vxlan_encap:
            d["gre"].update({'vxlan_tunnel_port': VXLAN_TUNNEL})
        return d

    def get_lg_ovs_flows(self, path_prefix, cookie_spec=None):
        output = {}
        for (table_name, table_id) in self.all_tables.items():
            output.update({
                "%s (%d)" % (table_name, table_id): self._run_command(
                    "ovs-ofctl -O %s dump-flows --names %s '%s' %s" % (
                        ovs_const.OPENFLOW15,
                        self.bridge.br_name,
                        dataplane_utils.join_s("table=%d" % table_id,
                                               cookie_spec),
                        OVS_DUMP_FLOW_FILTER),
                    run_as_root=True, shell=True
                )[0]
            })
        return output

    def get_lg_ovs_ports(self, path_prefix):
        (output, _) = self._run_command(
            "ovs-ofctl -O %s show %s |grep addr" % (ovs_const.OPENFLOW15,
                                                    self.bridge.br_name),
            run_as_root=True,
            acceptable_return_codes=[0, 1],
            shell=True)
        # FIXME: does it properly show the GRE tunnel interface ?
        return output



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from networking_bagpipe.agent.bagpipe_ml2 import agent_extension
from networking_bagpipe.bagpipe_bgp import constants as bbgp_const
from networking_bagpipe.tests.unit.agent import base


class TestBaGPipeAgentExtensionMixin(object):

    def _rts_for_network(self, network):
        rt = "64512:%s" % network['segmentation_id']
        return {
            "import_rt": [rt],
            "export_rt": [rt]
        }

    def test_bagpipe_l2_unsupported_network_type(self):
        # Set port network type to unsupported flat type
        port10_flat_net = self._port_data(base.PORT10)
        port10_flat_net['network_type'] = 'flat'

        self.agent_ext.handle_port(None, port10_flat_net)

        self.assertFalse(self.mocked_bagpipe_agent.do_port_plug.called)

        self._check_network_info(base.NETWORK1['id'], 0)

    def test_bagpipe_l2_attach_single_port(self):

        def check_build_cb(*args):
            # Verify build callback attachments
            local_port = self._get_expected_local_port(
                bbgp_const.EVPN, base.NETWORK1['id'],
                base.NETWORK1['segmentation_id'], base.PORT10['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    evpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        local_port=local_port['local_port'],
                        linuxbr=local_port['linuxbr'],
                        vni=base.NETWORK1['segmentation_id'],
                        **self._rts_for_network(base.NETWORK1)
                    )]
                ),
                self.agent_ext.build_bagpipe_l2_attach_info(base.PORT10['id'])
            )

        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'], 1)

    def test_bagpipe_l2_attach_multiple_ports_same_network(self):

        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self._check_network_info(base.NETWORK1['id'], 1)

        def check_build_cb(*args):
            for port in [base.PORT10, base.PORT11]:
                local_port = self._get_expected_local_port(
                    bbgp_const.EVPN, base.NETWORK1['id'],
                    base.NETWORK1['segmentation_id'], port['id'])
                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK1['id'],
                        evpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            local_port=local_port['local_port'],
                            linuxbr=local_port['linuxbr'],
                            vni=base.NETWORK1['segmentation_id'],
                            **self._rts_for_network(base.NETWORK1)
                        )]
                    ),
                    self.agent_ext.build_bagpipe_l2_attach_info(port['id'])
                )

        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self.agent_ext.handle_port(None, self._port_data(base.PORT11))
        self._check_network_info(base.NETWORK1['id'], 2)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])]
        )

    def test_bagpipe_l2_detach_single_port(self):

        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self._check_network_info(base.NETWORK1['id'], 1)

        def check_build_cb(*args):
            # Verify build callback attachments
            self.assertDictEqual(
                {},
                self.agent_ext.build_bagpipe_l2_attach_info(base.PORT10['id'])
            )

        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self.agent_ext.delete_port(None, self._port_data(base.PORT10))

        self._check_network_info(base.NETWORK1['id'], 0)

        local_port = self._get_expected_local_port(
            bbgp_const.EVPN, base.NETWORK1['id'],
            base.NETWORK1['segmentation_id'], base.PORT10['id'])
        detach_info = {
            'network_id': base.NETWORK1['id'],
            bbgp_const.EVPN: {
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': local_port['local_port']
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )


class LinuxBridgeAgentExtensionTest(base.BaseTestLinuxBridgeAgentExtension,
                                    TestBaGPipeAgentExtensionMixin):

    agent_extension_class = agent_extension.BagpipeML2AgentExtension



# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_bagpipe.bagpipe_bgp.engine import exa


@exa.NLRI.register(exa.AFI.ipv4, exa.SAFI.flow_vpn, force=True)
@exa.NLRI.register(exa.AFI.ipv6, exa.SAFI.flow_vpn, force=True)
class Flow(exa.Flow):
    '''ExaBGP Flow wrapper

    This wraps an ExaBGP Flow so that __eq__ and __hash__
    meet the criteria for RouteTableManager (in particular,
    not look at actions and nexthop)
    '''

    def __eq__(self, other):
        return self.pack() == other.pack()

    def __hash__(self):
        return hash(self.pack())

    def __repr__(self):
        return str(self)


def FlowRouteFactory(afi, rd):
    flow_route = Flow(afi, safi=exa.SAFI.flow_vpn)
    flow_route.rd = rd
    return flow_route



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from neutron_lib import constants
from oslo_utils import uuidutils
import testscenarios
import unittest

from neutron.tests.fullstack.resources import machine

from networking_bagpipe.tests.fullstack import base
from networking_bagpipe.tests.fullstack.resources.bgpvpn \
    import config as bgpvpn_cfg

load_tests = testscenarios.load_tests_apply_scenarios


class TestConnectivitySameBGPVPN(base.BaGPipeBaseFullStackTestCase):

    bagpipe_ml2 = False
    service_plugins = 'router,%s' % bgpvpn_cfg.BGPVPN_SERVICE

    of_interface = 'ovs-ofctl'
    bgpvpn = True

    port_per_compute_per_net = 2
    compute_node_count = 2

    scenarios = [
        ('OpenVSwitch MPLS-over-TEB-over-GRE', {
            'mech_drivers': 'openvswitch',
            'l2_agent_type': constants.AGENT_TYPE_OVS,
            'ipvpn_driver': 'ovs',
            'ipvpn_encap': 'mpls-gre'
        }),
        ('OpenVSwitch MPLS-over-GRE', {
            'mech_drivers': 'openvswitch',
            'l2_agent_type': constants.AGENT_TYPE_OVS,
            'ipvpn_driver': 'ovs',
            'ipvpn_encap': 'mpls-gre-l3'
        }),
        ('OpenVSwitch bare MPLS', {
            'mech_drivers': 'openvswitch',
            'l2_agent_type': constants.AGENT_TYPE_OVS,
            'ipvpn_driver': 'ovs',
            'ipvpn_encap': 'bare-mpls'
        }),
        ('Linuxbridge', {
            'mech_drivers': 'linuxbridge',
            'l2_agent_type': constants.AGENT_TYPE_LINUXBRIDGE,
            'ipvpn_driver': 'linux',
            'evpn_driver': 'linux',
            'ipvpn_encap': 'bare-mpls',
        })
    ]

    def test_l3_network_connectivity(self):
        tenant_uuid = uuidutils.generate_uuid()

        bgpvpn = self.safe_client.create_bgpvpn(tenant_uuid,
                                                route_targets=['64512:1'])

        network_ids = list()
        for subnet_cidr in (base.SUBNET_CIDR1, base.SUBNET_CIDR2):
            network_ids.append(
                self._create_net_subnet_bgpvpn_assoc(tenant_uuid, subnet_cidr,
                                                     bgpvpn['id'])[0]
            )

        fake_machines = list()
        for network_id in network_ids:
            fake_machines.extend([
                self.useFixture(
                    machine.FakeFullstackMachine(
                        self.environment.hosts[i],
                        network_id,
                        tenant_uuid,
                        self.safe_client))
                for i in
                range(
                    self.compute_node_count) * self.port_per_compute_per_net])

        vms = machine.FakeFullstackMachinesList(fake_machines)

        vms.block_until_all_boot()
        vms.ping_all()

    def test_l3_router_connectivity(self):
        tenant_uuid = uuidutils.generate_uuid()

        bgpvpn = self.safe_client.create_bgpvpn(tenant_uuid,
                                                route_targets=['64512:1'])

        network1 = self.safe_client.create_network(tenant_uuid)
        subnet1 = self.safe_client.create_subnet(
            tenant_uuid, network1['id'], '10.0.0.0/24')

        network2 = self.safe_client.create_network(tenant_uuid)
        subnet2 = self.safe_client.create_subnet(
            tenant_uuid, network2['id'], '20.0.0.0/24')

        router = self.safe_client.create_router(tenant_uuid)
        self.safe_client.add_router_interface(router['id'], subnet1['id'])
        self.safe_client.add_router_interface(router['id'], subnet2['id'])

        self.safe_client.create_router_association(tenant_uuid,
                                                   bgpvpn['id'],
                                                   router['id'])

        network3 = self.safe_client.create_network(tenant_uuid)
        self.safe_client.create_subnet(
            tenant_uuid, network3['id'], '30.0.0.0/24')
        self.safe_client.create_network_association(tenant_uuid,
                                                    bgpvpn['id'],
                                                    network3['id'])

        fake_machines = list()
        for network in (network1, network2, network3):
            fake_machines.extend([
                self.useFixture(
                    machine.FakeFullstackMachine(
                        self.environment.hosts[i],
                        network['id'],
                        tenant_uuid,
                        self.safe_client))
                for i in
                range(
                    self.compute_node_count) * self.port_per_compute_per_net])

        vms = machine.FakeFullstackMachinesList(fake_machines)

        vms.block_until_all_boot()
        vms.ping_all()

    @unittest.skip("Disabled because of bug 1715660 ( https://"
                   "bugs.launchpad.net/networking-bagpipe/+bug/1715660 )")
    def test_l2_network_connectivity(self):
        # create <n> fake machines in 2 different networks, all using
        # the same IP subnet, and check that each machine can reach all the
        # others. We create machines so that we confirm that connectivity
        # still works *inside* a given network, both locally on a compute
        # node, and across different compute nodes

        if self.evpn_driver == 'dummy':
            self.skipTest("L2VPN unsupported for this scenario")

        tenant_uuid = uuidutils.generate_uuid()

        bgpvpn = self.safe_client.create_bgpvpn(tenant_uuid,
                                                type="l2",
                                                route_targets=['64512:10'])

        fake_machines = list()
        for network in range(2):

            # we'll use the same subnet range for all networks, but
            # choose in this range distinct IP addresses for each fake machine

            network_id, subnet_id = self._create_net_subnet_bgpvpn_assoc(
                tenant_uuid,
                base.SUBNET_CIDR1,
                bgpvpn['id']
            )

            for compute, port_i in itertools.product(
                    range(self.compute_node_count),
                    range(self.port_per_compute_per_net)):

                # NOTE(tmorin): choice of fixed IP done this way for sake
                # of simplicity, of course, this breaks e.g. for
                # compute_node_count > 10
                fixed_ip = (
                        base.SUBNET_CIDR1[:base.SUBNET_CIDR1.find('0/24')] +
                        str(100 * network + 10 * (compute + 1) + port_i))

                neutron_port = self.safe_client.create_port(
                    network_id=network_id,
                    tenant_id=tenant_uuid,
                    hostname=self.environment.hosts[compute].hostname,
                    fixed_ips=[{"subnet_id": subnet_id,
                                "ip_address": fixed_ip}]
                )

                fake_machines.append(
                    self.useFixture(
                        machine.FakeFullstackMachine(
                            self.environment.hosts[compute],
                            network_id,
                            tenant_uuid,
                            self.safe_client,
                            neutron_port=neutron_port
                        )
                    )
                )

        vms = machine.FakeFullstackMachinesList(fake_machines)

        vms.block_until_all_boot()
        vms.ping_all()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2017 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This module is here to allow conciseness in exabgp imports
# by other modules

# flake8: noqa
from exabgp.bgp.message import OUT
from exabgp.bgp.message import IN

from exabgp.bgp.message.open.asn import ASN

from exabgp.bgp.message.update import Attributes
from exabgp.bgp.message.update.attribute.attribute import Attribute

from exabgp.bgp.message.update.attribute.community.extended.communities \
    import ExtendedCommunities
from exabgp.bgp.message.update.attribute.community.extended \
    import communities as extcoms
from exabgp.bgp.message.update.attribute.community.extended \
    import ConsistentHashSortOrder
from exabgp.bgp.message.update.attribute.community.extended \
    import Encapsulation
from exabgp.bgp.message.update.attribute.community.extended \
    import RouteTarget as RTExtCom
from exabgp.bgp.message.update.attribute.community.extended \
    import RouteTargetASN2Number as RouteTarget
from exabgp.bgp.message.update.attribute.community.extended \
    import TrafficRedirect
from exabgp.bgp.message.update.attribute.community.extended.rt_record\
    import RTRecord

from exabgp.bgp.message.update.attribute.localpref import LocalPreference
from exabgp.bgp.message.update.attribute.nexthop import NextHop
from exabgp.bgp.message.update.attribute.pmsi import PMSI
from exabgp.bgp.message.update.attribute.pmsi import PMSIIngressReplication

from exabgp.bgp.message.update.nlri.flow import Flow
from exabgp.bgp.message.update.nlri import flow
from exabgp.bgp.message.update.nlri.ipvpn import IPVPN
from exabgp.bgp.message.update.nlri.nlri import NLRI

from exabgp.bgp.message.update.nlri.qualifier.esi import ESI
from exabgp.bgp.message.update.nlri.qualifier.etag import EthernetTag
from exabgp.bgp.message.update.nlri.qualifier.labels import Labels
from exabgp.bgp.message.update.nlri.qualifier.mac import MAC
from exabgp.bgp.message.update.nlri.qualifier.rd import RouteDistinguisher

from exabgp.bgp.message.update.nlri.rtc import RTC

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN
from exabgp.bgp.message.update.nlri.evpn.mac import MAC as EVPNMAC
from exabgp.bgp.message.update.nlri.evpn.multicast import \
    Multicast as EVPNMulticast

from exabgp.protocol.ip import IP
from exabgp.protocol import Protocol

from exabgp.reactor.protocol import AFI
from exabgp.reactor.protocol import SAFI



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
L2 Agent extension to support bagpipe networking-sfc driver RPCs in the
Linux Bridge agent
"""
from copy import deepcopy

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_serialization import jsonutils

from oslo_concurrency import lockutils

from neutron_lib.agent import l2_extension
from neutron_lib import constants as n_const

from networking_bagpipe.agent import agent_base_info
from networking_bagpipe.agent import bagpipe_bgp_agent
from networking_bagpipe.bagpipe_bgp import constants as bbgp_const
from networking_bagpipe.driver import constants as sfc_const
from networking_bagpipe.objects import sfc as sfc_obj

from neutron.api.rpc.callbacks.consumer import registry as cons_registry
from neutron.api.rpc.callbacks import events as rpc_events
from neutron.api.rpc.handlers import resources_rpc

from neutron.plugins.ml2.drivers.linuxbridge.agent.common \
    import constants as lb_agt_constants
from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager

LOG = logging.getLogger(__name__)

SFC_SERVICE = 'sfc'


class BagpipeSfcAgentExtension(l2_extension.L2AgentExtension,
                               agent_base_info.BaseInfoManager):

    def initialize(self, connection, driver_type):

        if driver_type != lb_agt_constants.EXTENSION_DRIVER_TYPE:
            raise Exception("This extension is currently working only with the"
                            " Linux Bridge Agent")

        self.bagpipe_bgp_agent = (
            bagpipe_bgp_agent.BaGPipeBGPAgent.get_instance(
                n_const.AGENT_TYPE_LINUXBRIDGE)
        )

        self.bagpipe_bgp_agent.register_build_callback(
            SFC_SERVICE,
            self.build_sfc_attach_info)

        self.ports = set()
        self.bagpipe_bgp_agent.register_port_list(SFC_SERVICE,
                                                  self.ports)

        self._setup_rpc(connection)

    def _setup_rpc(self, connection):
        self._pull_rpc = resources_rpc.ResourcesPullRpcApi()
        self._register_rpc_consumers(connection)

    def _register_rpc_consumers(self, connection):
        endpoints = [resources_rpc.ResourcesPushRpcCallback()]

        # Consume BaGPipeChainHop OVO RPC
        cons_registry.register(self.handle_sfc_chain_hops,
                               sfc_obj.BaGPipeChainHop.obj_name())
        topic_chain_hop = resources_rpc.resource_type_versioned_topic(
            sfc_obj.BaGPipeChainHop.obj_name())
        connection.create_consumer(topic_chain_hop, endpoints, fanout=True)

        # Consume BaGPipePortHops OVO RPC
        cons_registry.register(self.handle_sfc_port_hops,
                               sfc_obj.BaGPipePortHops.obj_name())
        topic_port_hops = resources_rpc.resource_type_versioned_topic(
            sfc_obj.BaGPipePortHops.obj_name())
        connection.create_consumer(topic_port_hops, endpoints, fanout=True)

    def _add_sfc_chain_hop_helper_for_port(self, port_id, chain_hop, side,
                                           lb_consistent_hash_order):
        orig_info = deepcopy(chain_hop)
        port_info = self.ports_info[port_id]

        # Set gateway IP address in NetworkInfo if necessary
        net_info = port_info.network
        if net_info.gateway_info == agent_base_info.NO_GW_INFO:
            gateway_info = (
                agent_base_info.GatewayInfo(None,
                                            chain_hop[side + '_gw'])
            )
            net_info.set_gateway_info(gateway_info)

        # FIXME: We could use subset bytes of port pair id to avoid hash
        # polarization
        # Chain hop ingress, respectively egress, port corresponds to SFC port
        # pair egress, respectively ingress, port side.
        if side == sfc_const.EGRESS:
            orig_info['lb_consistent_hash_order'] = (
                lb_consistent_hash_order
            )

        port_info.add_chain_hop({side: orig_info})

    def _add_sfc_chain_hop_helper(self, port_ids, chain_hop, side):
        for index, port_id in enumerate(port_ids):
            if port_id not in self.ports_info:
                continue

            self._add_sfc_chain_hop_helper_for_port(port_id, chain_hop, side,
                                                    index)

    def _remove_sfc_chain_hop_helper(self, port_info, chain_hop, side):
        sfc_info = port_info.chain_hops
        if side in sfc_info:
            if side == sfc_const.EGRESS:
                sfc_info[side].pop('lb_consistent_hash_order')

            if chain_hop != sfc_info[side]:
                LOG.warning("%s service inconsistent %s informations for "
                            "port %s", SFC_SERVICE, side, port_info.id)
                return

            sfc_info.pop(side)

    def build_sfc_attach_info(self, port_id):
        if port_id not in self.ports_info:
            LOG.warning("%s service has no PortInfo for port %s",
                        SFC_SERVICE, port_id)
            return {}

        port_info = self.ports_info[port_id]

        ipvpn_attachment = self._build_attachment(port_info)

        attachments = {}

        if ipvpn_attachment:
            attachments[bbgp_const.IPVPN] = [ipvpn_attachment]
            attachments['network_id'] = port_info.network.id

        return attachments

    def _build_attachment(self, port_info):
        if not port_info.chain_hops:
            return {}

        linuxbr = LinuxBridgeManager.get_bridge_name(port_info.network.id)
        chain_hop = port_info.chain_hops

        attachment = {
            'ip_address': port_info.ip_address,
            'mac_address': port_info.mac_address,
            'gateway_ip': port_info.network.gateway_info.ip,
            'local_port': {'linuxif': linuxbr},
            bbgp_const.RT_IMPORT: (chain_hop[sfc_const.INGRESS]['rts']
                                   if chain_hop.get(sfc_const.INGRESS)
                                   else []),
            bbgp_const.RT_EXPORT: (chain_hop[sfc_const.EGRESS]['rts']
                                   if chain_hop.get(sfc_const.EGRESS)
                                   else [])
        }

        egress_info = chain_hop.get(sfc_const.EGRESS)
        if egress_info:
            if egress_info.get('readv_to_rt'):
                attachment.update({
                    'readvertise': dict(
                        from_rt=egress_info.get('readv_from_rts', []),
                        to_rt=[egress_info['readv_to_rt']]
                    )
                })

            if (egress_info.get('redirect_rts') and
                    egress_info.get('classifiers')):
                classifier = jsonutils.loads(egress_info['classifiers'])[0]

                attract_traffic = dict(
                    redirect_rts=egress_info['redirect_rts'],
                    classifier=classifier
                )

                if egress_info.get('attract_to_rt'):
                    destination_prefix = classifier.get('destinationPrefix',
                                                        '0.0.0.0/0')

                    attract_traffic.update(dict(
                        to_rt=[egress_info['attract_to_rt']],
                        static_destination_prefixes=[destination_prefix])
                    )

                attachment.update(
                    dict(attract_traffic=attract_traffic)
                )

            attachment.update(dict(
                lb_consistent_hash_order=egress_info[
                    'lb_consistent_hash_order'])
            )

        return attachment

    def _build_sfc_detach_info(self, port_info):
        linuxbr = LinuxBridgeManager.get_bridge_name(port_info.network.id)

        detach_info = {
            'network_id': port_info.network.id,
            bbgp_const.IPVPN: {
                'ip_address': port_info.ip_address,
                'mac_address': port_info.mac_address,
                'local_port': {'linuxif': linuxbr}
            }
        }

        return detach_info

    def handle_sfc_chain_hops(self, context, resource_type, chain_hops,
                              event_type):
        LOG.debug("handle_sfc_chain_hops called with: {resource_type: %s, "
                  "chain_hops: %s, event_type: %s",
                  resource_type, chain_hops, event_type)
        if event_type == rpc_events.CREATED:
            self.sfc_ports_attach(chain_hops)
        elif event_type == rpc_events.DELETED:
            self.sfc_ports_detach(chain_hops)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-sfc')
    def sfc_ports_attach(self, chain_hops):
        attach_ids = set()

        for hop in chain_hops:
            hop_dict = hop.to_dict()

            ingress_ids = hop_dict.pop('ingress_ports')
            egress_ids = hop_dict.pop('egress_ports')

            if ingress_ids:
                self._add_sfc_chain_hop_helper(ingress_ids, hop_dict,
                                               sfc_const.INGRESS)
                attach_ids.update(ingress_ids)

            if egress_ids:
                self._add_sfc_chain_hop_helper(egress_ids, hop_dict,
                                               sfc_const.EGRESS)
                attach_ids.update(egress_ids)

        port_ids = [p_id for p_id in attach_ids if p_id in self.ports_info]
        for port_id in port_ids:
            self.bagpipe_bgp_agent.do_port_plug(port_id)

    def _remove_sfc_info_for_port(self, port_id, side, chain_hop,
                                  ports_to_detach):
        port_info = self.ports_info.get(port_id)
        if not port_info:
            LOG.warning("%s service inconsistent for port %s",
                        SFC_SERVICE, port_id)
            return

        if port_id not in ports_to_detach:
            detach_info = self._build_sfc_detach_info(port_info)

            ports_to_detach[port_id] = detach_info

        self._remove_sfc_chain_hop_helper(port_info, chain_hop, side)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-sfc')
    def sfc_ports_detach(self, chain_hops):
        ports_to_detach = dict()

        for hop in chain_hops:
            hop_dict = hop.to_dict()

            ingress_ids = hop_dict.pop('ingress_ports')
            egress_ids = hop_dict.pop('egress_ports')

            if ingress_ids:
                for port_id in ingress_ids:
                    self._remove_sfc_info_for_port(port_id,
                                                   sfc_const.INGRESS,
                                                   hop_dict,
                                                   ports_to_detach)

            if egress_ids:
                for port_id in egress_ids:
                    self._remove_sfc_info_for_port(port_id,
                                                   sfc_const.EGRESS,
                                                   hop_dict,
                                                   ports_to_detach)

        for port_id, detach_info in ports_to_detach.items():
            self.bagpipe_bgp_agent.do_port_plug_refresh(port_id, detach_info)

    def handle_sfc_port_hops(self, context, resource_type, port_hops,
                             event_type):
        LOG.debug("handle_sfc_port_hops called with: {resource_type: %s, "
                  "port_hops: %s, event_type: %s",
                  (resource_type, port_hops, event_type))
        if event_type == rpc_events.CREATED:
            self.sfc_add_port_hops(port_hops)
        elif event_type == rpc_events.DELETED:
            self.sfc_remove_port_hops(port_hops)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-sfc')
    def sfc_add_port_hops(self, port_hops):
        for port_hop in port_hops:
            port_id = port_hop.port_id
            for side in [sfc_const.INGRESS, sfc_const.EGRESS]:
                for hop in getattr(port_hop, side + '_hops'):
                    hop_dict = hop.to_dict()

                    lb_consistent_hash_order = (
                        hop_dict[side + '_ports'].index(port_id)
                    )

                    del hop_dict['ingress_ports']
                    del hop_dict['egress_ports']

                    self._add_sfc_chain_hop_helper_for_port(
                        port_id, hop_dict, side, lb_consistent_hash_order
                    )

            self.bagpipe_bgp_agent.do_port_plug(port_id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-sfc')
    def sfc_remove_port_hops(self, port_hops):
        for port_hop in port_hops:
            port_id = port_hop.port_id
            port_info = self.ports_info.get(port_id)

            if not port_info:
                LOG.warning("%s service inconsistent for port %s",
                            SFC_SERVICE, port_id)
                continue

            detach_info = self._build_sfc_detach_info(port_info)
            self.ports_info[port_id].chain_hops = {}

            self.bagpipe_bgp_agent.do_port_plug_refresh(port_id, detach_info)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-sfc')
    def handle_port(self, context, port):
        port_id = port['port_id']
        net_id = port['network_id']

        self.ports.add(port_id)

        net_info, port_info = (
            self._get_network_port_infos(net_id, port_id)
        )

        port_info.mac_address = port['mac_address']
        port_info.ip_address = port['fixed_ips'][0]['ip_address']

        port_hops = self._pull_rpc.bulk_pull(
            context,
            sfc_obj.BaGPipeChainHop.obj_name(),
            filter_kwargs=dict(port_id=port_id))

        for port_hop in port_hops:
            hop_dict = port_hop.to_dict()
            ingress_ports = hop_dict.pop('ingress_ports')
            egress_ports = hop_dict.pop('egress_ports')

            if port_id in ingress_ports:
                self._add_sfc_chain_hop_helper_for_port(
                    port_id, hop_dict, sfc_const.INGRESS,
                    ingress_ports.index(port_id)
                )

            if port_id in egress_ports:
                self._add_sfc_chain_hop_helper_for_port(
                    port_id, hop_dict, sfc_const.EGRESS,
                    egress_ports.index(port_id)
                )

        if port_info.chain_hops:
            self.bagpipe_bgp_agent.do_port_plug(port_id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-sfc')
    def delete_port(self, context, port):
        port_id = port['port_id']
        port_info = self.ports_info.get(port_id)

        if port_info and port_info.chain_hops:
            detach_info = self._build_sfc_detach_info(port_info)

            self._remove_network_port_infos(port_info.network.id, port_id)
            self.ports.remove(port_id)

            self.bagpipe_bgp_agent.do_port_plug_refresh(port_id,
                                                        detach_info)



# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from pyroute2 import ndb as ndb_mod  # pylint: disable=no-name-in-module

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp import constants as consts
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn import evpn
from networking_bagpipe.privileged import privileged_utils

BRIDGE_NAME_PREFIX = "evpn---"
VXLAN_INTERFACE_PREFIX = "vxlan--"
LOG = logging.getLogger(__name__)


class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):

    def __init__(self, *args, **kwargs):
        super(LinuxVXLANEVIDataplane, self).__init__(*args, **kwargs)

        if 'linuxbr' in kwargs:
            self.bridge_name = kwargs.get('linuxbr')
        else:
            self.bridge_name = (
                BRIDGE_NAME_PREFIX +
                self.external_instance_id)[:consts.LINUX_DEV_LEN]

        if not self._interface_exists(self.bridge_name):
            self.log.debug("Starting bridge %s", self.bridge_name)

            # Create bridge
            privileged_utils.brctl('addbr %s' % self.bridge_name)
            privileged_utils.brctl('setfd %s 0' % self.bridge_name)
            privileged_utils.brctl('stp %s off' % self.bridge_name)
            self._run_command("ip link set %s up" % self.bridge_name,
                              run_as_root=True)

            self.log.debug("Bridge %s created", self.bridge_name)

        self._create_and_plug_vxlan_if()

        self.log.debug("VXLAN interface %s plugged on bridge %s",
                       self.vxlan_if_name, self.bridge_name)

        self._cleaning_up = False

    @log_decorator.log_info
    def cleanup(self):
        self.log.info("Cleaning EVI bridge and VXLAN interface %s",
                      self.bridge_name)

        self._cleaning_up = True

        # removing the vxlan interface removes our routes,
        # but if we don't remove the vxlan if (if it was reused) then
        # cleanup will not happen, which is why we use cleanup assist
        # (see needs cleanup assist below)
        self._cleanup_vxlan_if()

        # Delete only EVPN Bridge (Created by dataplane driver)
        if BRIDGE_NAME_PREFIX in self.bridge_name:
            self._run_command("ip link set %s down" % self.bridge_name,
                              run_as_root=True,
                              raise_on_error=False)
            privileged_utils.brctl('delbr %s' % self.bridge_name,
                                   check_exit=False)

    def needs_cleanup_assist(self):
        # If we reused a vxlan interface we won't cleanup fdb entries
        # in cleanup(), so we need to have remove_dataplane_for_x
        # be called for reach state via cleanup assist
        return VXLAN_INTERFACE_PREFIX not in self.vxlan_if_name

    def _create_and_plug_vxlan_if(self):
        # if a VXLAN interface, with the VNI we want to use, is already plugged
        # in the bridge, we want to reuse it
        with ndb_mod.main.NDB() as ndb:
            # pylint: disable=no-member
            for port_id in ndb.interfaces[self.bridge_name].ports:
                port = ndb.interfaces[port_id]  # pylint: disable=no-member
                if (port['kind'] == "vxlan" and
                        port['vxlan_id'] == self.instance_label):
                    self.log.info("reuse vxlan interface %s for VXLAN VNI %s",
                                  port['ifname'], self.instance_label)
                    self.vxlan_if_name = port['ifname']
                    return

        self.vxlan_if_name = (VXLAN_INTERFACE_PREFIX +
                              self.external_instance_id)[:consts.LINUX_DEV_LEN]

        self.log.debug("Creating and plugging VXLAN interface %s",
                       self.vxlan_if_name)

        if self._interface_exists(self.vxlan_if_name):
            self._remove_vxlan_if()

        dst_port_spec = ""
        if self.driver.config.vxlan_dst_port:
            dst_port_spec = ("dstport %d" %
                             self.driver.config.vxlan_dst_port)

        # Create VXLAN interface
        self._run_command(
            "ip link add %s type vxlan id %d local %s nolearning proxy %s" %
            (self.vxlan_if_name, self.instance_label,
             self.driver.get_local_address(), dst_port_spec),
            run_as_root=True
        )

        self._run_command("ip link set %s up" % self.vxlan_if_name,
                          run_as_root=True)

        # Plug VXLAN interface into bridge
        privileged_utils.brctl('addif %s %s' % (self.bridge_name,
                                                self.vxlan_if_name))

    def _cleanup_vxlan_if(self):
        if VXLAN_INTERFACE_PREFIX not in self.vxlan_if_name:
            self.log.debug("we reused the VXLAN interface, don't cleanup")
            return

        if self._is_vxlan_if_on_bridge():
            # Unplug VXLAN interface from Linux bridge
            self._unplug_from_bridge(self.vxlan_if_name)

        self._remove_vxlan_if()

    def _remove_vxlan_if(self):
        if not VXLAN_INTERFACE_PREFIX not in self.vxlan_if_name:
            self.log.debug("we reused the VXLAN interface, don't remove")
            return

        # Remove VXLAN interface
        self._run_command("ip link set %s down" % self.vxlan_if_name,
                          run_as_root=True)
        self._run_command("ip link del %s" % self.vxlan_if_name,
                          run_as_root=True)

    def _is_if_on_bridge(self, ifname):
        with ndb_mod.main.NDB() as ndb:
            try:
                # pylint: disable=no-member
                for port_id in ndb.interfaces[self.bridge_name].ports:
                    port = ndb.interfaces[port_id]  # pylint: disable=no-member
                    if port.ifname == ifname:
                        return True
            except KeyError:
                return False
        return False

    def _is_vxlan_if_on_bridge(self):
        return self._is_if_on_bridge(self.vxlan_if_name)

    def _interface_exists(self, interface):
        """Check if interface exists."""
        (_, exit_code) = self._run_command("ip link show dev %s" % interface,
                                           raise_on_error=False,
                                           acceptable_return_codes=[-1])
        return (exit_code == 0)

    def _unplug_from_bridge(self, interface):
        if self._interface_exists(self.bridge_name):
            privileged_utils.brctl('delif %s %s' % (self.bridge_name,
                                                    interface),
                                   check_exit=[0, 1])

    def set_gateway_port(self, linuxif, gw_ip):
        privileged_utils.brctl('addif %s %s' % (self.bridge_name, linuxif),
                               check_exit=False)

        self._fdb_dump()

    def gateway_port_down(self, linuxif):
        privileged_utils.brctl('delif %s %s' % (self.bridge_name,
                                                linuxif),
                               check_exit=False)
        # TODO(tmorin): need to cleanup bridge fdb and ip neigh ?

    def set_bridge_name(self, linuxbr):
        self.bridge_name = linuxbr

    @log_decorator.log_info
    def vif_plugged(self, mac_address, ip_address, localport, dpid, direction):
        # Plug localport only if bridge was created by us
        if BRIDGE_NAME_PREFIX in self.bridge_name:
            self.log.debug("Plugging localport %s into EVPN bridge %s",
                           localport['linuxif'], self.bridge_name)
            privileged_utils.brctl(
                'addif %s %s' % (self.bridge_name, localport['linuxif']),
                check_exit=False)

        privileged_utils.bridge(
            'fdb replace %s dev %s' % (mac_address, localport['linuxif']))

        self._fdb_dump()

    @log_decorator.log_info
    def vif_unplugged(self, mac_address, ip_address, localport, dpid,
                      direction, last_endpoint=True):

        # remove local fdb entry, but only if tap interface is still here
        if self._is_if_on_bridge(localport['linuxif']):
            privileged_utils.bridge(
                'fdb delete %s dev %s' % (mac_address,
                                          localport['linuxif']))

        # unplug localport only if bridge was created by us
        if BRIDGE_NAME_PREFIX in self.bridge_name:
            self.log.debug("Unplugging localport %s from EVPN bridge %s",
                           localport['linuxif'], self.bridge_name)
            self._unplug_from_bridge(localport['linuxif'])

        self._fdb_dump()

    @log_decorator.log
    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, dpid,
                                            nlri, encaps):
        if self._cleaning_up:
            self.log.debug("setup_dataplane_for_remote_endpoint: instance"
                           " cleaning up, do nothing")
            return

        mac = prefix
        ip = nlri.ip
        vni = dpid

        # populate bridge forwarding db
        privileged_utils.bridge(
            'fdb replace %s dev %s dst %s vni %s' %
            (mac, self.vxlan_if_name, remote_pe, vni))

        # populate ARP cache
        if ip is not None:
            self._run_command("ip neighbor replace %s lladdr %s dev %s nud "
                              "permanent" % (ip, mac, self.vxlan_if_name),
                              run_as_root=True)
        else:
            self.log.trace("No IP in E-VPN route, no ARP proxy for %s" % mac)

        self._fdb_dump()

    @log_decorator.log
    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, dpid,
                                             nlri):
        if self._cleaning_up:
            self.log.debug("setup_dataplane_for_remote_endpoint: instance"
                           " cleaning up, do nothing")
            return

        mac = prefix
        ip = nlri.ip
        vni = dpid

        self._fdb_dump()

        # clear ARP proxy
        if ip is not None:
            self._run_command("ip neighbor del %s lladdr %s dev %s nud "
                              "permanent" % (ip, mac, self.vxlan_if_name),
                              run_as_root=True)

        privileged_utils.bridge(
            'fdb del %s dev %s dst %s vni %s' %
            (mac, self.vxlan_if_name, remote_pe, vni))

        self._fdb_dump()

    @log_decorator.log
    def add_dataplane_for_bum_endpoint(self, remote_pe, dpid, nlri, encaps):
        if self._cleaning_up:
            self.log.debug("setup_dataplane_for_remote_endpoint: instance"
                           " cleaning up, do nothing")
            return

        vni = dpid

        # 00:00:00:00:00 usable as default since kernel commit
        # 58e4c767046a35f11a55af6ce946054ddf4a8580 (2013-06-25)
        privileged_utils.bridge(
            'fdb append 00:00:00:00:00:00 dev %s dst %s vni %s' %
            (self.vxlan_if_name, remote_pe, vni))

        self._fdb_dump()

    @log_decorator.log
    def remove_dataplane_for_bum_endpoint(self, remote_pe, dpid, nlri):
        if self._cleaning_up:
            self.log.debug("setup_dataplane_for_remote_endpoint: instance"
                           " cleaning up, do nothing")
            return

        vni = dpid

        self._fdb_dump()

        privileged_utils.bridge(
            'fdb delete 00:00:00:00:00:00 dev %s dst %s vni %s' %
            (self.vxlan_if_name, remote_pe, vni))

        self._fdb_dump()

    def _fdb_dump(self):
        if self.log.isEnabledFor(logging.DEBUG):
            self.log.debug("bridge fdb dump: %s", self._run_command(
                "fdb show br %s" % self.bridge_name,
                acceptable_return_codes=[0, 255],
                run_as_root=True)[0])

    # Looking glass ####

    def get_lg_local_info(self, path_prefix):
        return {
            "linux_bridge": self.bridge_name,
            "vxlan_if": self.vxlan_if_name
        }


class LinuxVXLANDataplaneDriver(dp_drivers.DataplaneDriver):

    """E-VPN Dataplane driver relying on Linux kernel linuxbridge VXLAN"""

    dataplane_instance_class = LinuxVXLANEVIDataplane
    type = consts.EVPN
    required_kernel = "3.11.0"
    encaps = [exa.Encapsulation(exa.Encapsulation.Type.VXLAN)]

    driver_opts = [
        cfg.IntOpt("vxlan_dst_port", default="4789",
                   help=("UDP port toward which send VXLAN traffic (defaults "
                         "to standard IANA-allocated port)")),
    ]

    def __init__(self):
        super(LinuxVXLANDataplaneDriver, self).__init__()

        privileged_utils.modprobe('vxlan')

    @log_decorator.log_info
    def reset_state(self):
        # delete all EVPN bridges
        res = privileged_utils.brctl('show', check_exit=[0, 1])
        if res and len(res) > 0:
            res = res[0]
        bridge_lines = [line for line in res.split('\n')[1:-1]
                        if BRIDGE_NAME_PREFIX in line]

        for b_l in bridge_lines:
            bridge = b_l.split('\t')[0]
            self._run_command("ip link set %s down" % bridge,
                              run_as_root=True)
            privileged_utils.brctl('delbr %s' % bridge)

        # delete all VXLAN interfaces
        cmd = "ip link show | awk '{print $2}' | tr -d ':' | grep '%s'"
        for interface in self._run_command(cmd % VXLAN_INTERFACE_PREFIX,
                                           run_as_root=True,
                                           raise_on_error=False,
                                           acceptable_return_codes=[0, 1],
                                           shell=True)[0]:
            self._run_command("ip link set %s down" % interface,
                              run_as_root=True)
            self._run_command("ip link delete %s" % interface,
                              run_as_root=True)



# Copyright 2018 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock

from networking_bagpipe.bagpipe_bgp.common import dataplane_utils


class TestObjectLifecycleManager(testtools.TestCase):

    def setUp(self):
        super(TestObjectLifecycleManager, self).setUp()

        self.test_object_mgr = dataplane_utils.ObjectLifecycleManager()
        self.test_object_mgr.create_object = mock.Mock(return_value=1)
        self.test_object_mgr.delete_object = mock.Mock()

    def test_get_object_first_user(self):
        _, first1 = self.test_object_mgr.get_object("OBJ1", "USER_A")
        self.assertTrue(first1)

        self.test_object_mgr.create_object.assert_called_once()
        self.assertEqual(len(self.test_object_mgr.objects), 1)
        self.assertEqual(len(self.test_object_mgr.object_used_for["OBJ1"]), 1)

    def test_get_object_multiple_users(self):
        self.test_object_mgr.get_object("OBJ1", "USER_A")

        _, first2 = self.test_object_mgr.get_object("OBJ1", "USER_B")
        self.assertFalse(first2)

        self.test_object_mgr.create_object.assert_called_once()
        self.assertEqual(len(self.test_object_mgr.objects), 1)
        self.assertEqual(len(self.test_object_mgr.object_used_for["OBJ1"]), 2)

    def test_find_object_already_exist(self):
        self.test_object_mgr.get_object("OBJ1", "USER_A")

        obj1 = self.test_object_mgr.find_object("OBJ1")
        self.assertIsNotNone(obj1)

    def test_find_object_empty(self):
        obj1 = self.test_object_mgr.find_object("OBJ1")
        self.assertIsNone(obj1)

    def test_free_object_last_user(self):
        self.test_object_mgr.get_object("OBJ1", "USER_A")

        last1 = self.test_object_mgr.free_object("OBJ1", "USER_A")
        self.assertTrue(last1)

        self.test_object_mgr.delete_object.assert_called_once()
        self.assertTrue(not self.test_object_mgr.objects)
        self.assertTrue(not self.test_object_mgr.object_used_for)

    def test_free_object_multiple_users(self):
        self.test_object_mgr.get_object("OBJ1", "USER_A")
        self.test_object_mgr.get_object("OBJ1", "USER_B")

        last1 = self.test_object_mgr.free_object("OBJ1", "USER_A")
        self.assertFalse(last1)

        self.test_object_mgr.delete_object.assert_not_called()
        self.assertEqual(len(self.test_object_mgr.objects), 1)
        self.assertEqual(len(self.test_object_mgr.object_used_for["OBJ1"]), 1)

        last2 = self.test_object_mgr.free_object("OBJ1", "USER_B")
        self.assertTrue(last2)

        self.test_object_mgr.delete_object.assert_called_once()
        self.assertFalse(self.test_object_mgr.objects)
        self.assertFalse(self.test_object_mgr.object_used_for)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_messaging

from neutron_lib.agent import topics
from neutron_lib import rpc as n_rpc
from oslo_log import log as logging


LOG = logging.getLogger(__name__)

# until we have a better way to add something in the topic namespace
# from a python package external to Neutron...
topics_BAGPIPE_BGPVPN = "bagpipe-bgpvpn"


class BGPVPNAgentNotifyApi(object):
    """Base class for BGP VPN Service Plugin notification to agent RPC API."""

    def __init__(self, topic=topics.AGENT):
        self.topic = topic

        self.topic_bgpvpn_update = topics.get_topic_name(self.topic,
                                                         topics_BAGPIPE_BGPVPN,
                                                         topics.UPDATE)

        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    # BGP VPN CRUD notifications
    # --------------------------------------
    def _notification_fanout(self, context, method, bgpvpn):
        LOG.debug('Fanout notify BGP VPN agents at %(topic)s '
                  'the message %(method)s with %(bgpvpn)s',
                  {'topic': self.topic_bgpvpn_update,
                   'method': method,
                   'bgpvpn': bgpvpn})

        cctxt = self.client.prepare(topic=self.topic_bgpvpn_update,
                                    fanout=True)
        cctxt.cast(context, method, bgpvpn=bgpvpn)

    def create_bgpvpn(self, context, bgpvpn):
        return self._notification_fanout(context,
                                         'create_bgpvpn',
                                         bgpvpn)

    def update_bgpvpn(self, context, bgpvpn):
        return self._notification_fanout(context,
                                         'update_bgpvpn',
                                         bgpvpn)

    def delete_bgpvpn(self, context, bgpvpn):
        return self._notification_fanout(context,
                                         'delete_bgpvpn',
                                         bgpvpn)

    # Port attach/detach on/from BGP VPN notifications
    # ---------------------------------------------------------
    def _notification_host(self, context, method, port_bgpvpn_info, host):
        LOG.debug('Notify BGP VPN agent %(host)s at %(topic)s '
                  'the message %(method)s with %(port_bgpvpn_info)s',
                  {'host': host,
                   'topic': self.topic_bgpvpn_update,
                   'method': method,
                   'port_bgpvpn_info': port_bgpvpn_info})

        cctxt = self.client.prepare(topic=self.topic_bgpvpn_update,
                                    server=host)
        cctxt.cast(context, method, port_bgpvpn_info=port_bgpvpn_info)

    def attach_port_on_bgpvpn(self, context, port_bgpvpn_info, host=None):
        if port_bgpvpn_info:
            self._notification_host(context, 'attach_port_on_bgpvpn',
                                    port_bgpvpn_info, host)

    def detach_port_from_bgpvpn(self, context, port_bgpvpn_info, host=None):
        self._notification_host(context, 'detach_port_from_bgpvpn',
                                port_bgpvpn_info, host)



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_bagpipe.bagpipe_bgp import constants as bbgp_const

from neutron_lib.api.definitions import bgpvpn

BGPVPN_SERVICE = 'bgpvpn'

DEFAULT_GATEWAY_MAC = "00:00:5e:00:43:64"
FALLBACK_SRC_MAC = "00:00:5e:2a:10:00"

# Map from BGPVPN service VPN types to bagpipe-bgp VPN types
BGPVPN_2_BAGPIPE = {bgpvpn.BGPVPN_L2: bbgp_const.EVPN,
                    bgpvpn.BGPVPN_L3: bbgp_const.IPVPN}

LINUXIF_PREFIX = "patch2tun"
